==> luminarycloud-0.2.2-py3-none-any.whl: luminarycloud/_helpers/upload_mesh.py <==
# Copyright 2023 Luminary Cloud, Inc. All Rights Reserved.
import cgi
import io
import logging
import os
import hashlib
import requests
from tempfile import TemporaryDirectory
from typing import cast, Iterator, Tuple, Optional, Union
from urllib.parse import urlparse

from .._proto.common import common_pb2 as commonpb
from .._proto.mesh.mesh_pb2 import Mesh, UploadMeshRequest, UploadMeshResponse
from .._client import Client

logger = logging.getLogger(__name__)

DEFAULT_CHUNK_SIZE_BYTES = 1024 * 1024


def _chunker(
    stream: io.BufferedIOBase,
    chunk_size: int = None,
) -> Iterator[Tuple[int, bytes]]:
    """
    Generator which breaks the stream into chunks.

    Parameters
    ----------
    stream: io.BufferedIOBase
        The stream to chunk
    chunk_size: int
        (Optional) the default chunk size is 1MB

    Yields
    ------
    offset: int
    chunk: bytes
    """
    if chunk_size is None:
        chunk_size = DEFAULT_CHUNK_SIZE_BYTES

    offset = 0
    while True:
        chunk = stream.read(chunk_size)
        bytes_read = len(chunk)
        if bytes_read == 0:
            return
        yield offset, chunk
        if bytes_read < chunk_size:
            return
        offset += bytes_read


def _digest(
    path: Union[os.PathLike, str],
    *,
    chunk_size: int = None,
) -> bytes:
    """
    Computes the sha256 digest of the file at the given path.

    Uses a buffer to avoid loading the entire file into memory.

    Parameters
    ----------
    path: PathLike or str
        The absolute or relative file path
    chunk_size : int
        (Optional) the default is 1MB

    Returns
    -------
    digest: bytes
    """
    with open(path, "rb") as fp:
        m = hashlib.sha256()
        for _, chunk in _chunker(fp, chunk_size):
            m.update(chunk)
    return m.digest()


def _upload_mesh_from_stream(
    client: Client,
    binary_stream: io.BufferedIOBase,
    project_id: str,
    filename: str,
    size: int,
    sha256checksum: bytes,
    *,
    name: Optional[str] = None,
    scaling: Optional[float] = None,
    chunk_size: Optional[int] = None,
) -> UploadMeshResponse:
    """
    Upload mesh from a binary stream.

    Parameters
    ----------
    client: Client
        A LuminaryCloud Client (see client.py)
    binary_stream: io.BufferedIOBase
        A fresh read stream of the mesh file to upload
    project_id: str
        The ID of the project to upload the mesh under
    filename: str
        The filename of the mesh. Must include the appropriate filename extension.
    size: int
        The size of the stream to be read
    sha256checksum: str
        The digest of the mesh file to upload
    name: str
        (Optional) the name of the mesh resource on Luminary Cloud. defaults to the filename
        (without the extension).
    chunk_size: int
        (Optional) the default size is 1MB.
    """

    assert size > 0, "stream size must be positive"
    fparts = filename.split(".", maxsplit=1)
    if len(fparts) < 2:
        logger.error("Filename is missing extension.")
        raise ValueError("Filename is missing an extension.")
    mesh_name, file_ext = fparts
    if name is not None:
        mesh_name = cast(str, name)
    logger.debug(f"Uploading mesh with name '{mesh_name}' and extension '{file_ext}'.")

    logger.debug("Uploading chunks.")
    response = None
    next_upload_token = ""
    file_metadata = commonpb.FileMetadata(
        name=mesh_name, ext=file_ext, size=size, sha256_checksum=sha256checksum
    )
    for offset, chunk in _chunker(binary_stream, chunk_size):
        req = UploadMeshRequest(
            project_id=project_id,
            name=mesh_name,
            upload_token=next_upload_token,
            file_chunk=commonpb.FileChunk(
                metadata=file_metadata,
                bytes=commonpb.ByteChunk(
                    start_offset=offset,
                    data=chunk,
                ),
            ),
        )
        if scaling is not None:
            req.scaling = cast(float, scaling)
        response = client.UploadMesh(req)
        next_upload_token = response.upload_token

    if response is None:
        logger.error("Stream is empty.")
        raise ValueError("Empty stream")

    return cast(UploadMeshResponse, response)


def upload_mesh_from_local_file(
    client: Client,
    project_id: str,
    path: Union[os.PathLike, str],
    *,
    name: Optional[str] = None,
    scaling: Optional[float] = None,
    chunk_size: Optional[int] = None,
) -> Mesh:
    """
    Upload a mesh from a local file.

    The mesh file format is inferred from the filename extension.
    For supported formats, see: https://learning.luminarycloud.com/guides/mesh-file-formats

    Parameters
    ----------
    client: Client
        A LuminaryCloud Client (see client.py)
    project_id: str
        The ID of the project to upload the mesh under
    path: PathLike or str
        The relative or absolute file path of the mesh file to upload.
    name: str
        (Optional) the name of the mesh resource on Luminary Cloud. defaults to the filename
        (without the extension).
    chunk_size: int
        (Optional) the default size is 1MB.
    """
    if not os.path.exists(path):
        logger.error("File not found.")
        raise FileNotFoundError

    with open(path, "rb") as fp:
        response = _upload_mesh_from_stream(
            client,
            fp,
            project_id,
            os.path.basename(path),
            os.path.getsize(path),
            _digest(path),
            name=name,
            scaling=scaling,
            chunk_size=chunk_size,
        )

    return response.mesh


def upload_mesh_from_url(
    client: Client,
    project_id: str,
    url: str,
    *,
    timeout: int = 5,
    name: Optional[str] = None,
    scaling: Optional[float] = None,
    chunk_size: Optional[int] = None,
) -> Mesh:
    """
    Upload a mesh from a URL.

    The mesh file format is inferred from the filename extension.
    For supported formats, see: https://learning.luminarycloud.com/guides/mesh-file-formats

    Parameters
    ----------
    client: Client
        A LuminaryCloud Client (see client.py)
    project_id: str
        The ID of the project to upload the mesh under
    url: str
        The url of the mesh file.
    timeout: int
        (Optional) timeout (in seconds) for the download request.
    name: str
        (Optional) the name of the mesh resource on Luminary Cloud. defaults to the filename
        (without the extension).
    chunk_size: int
        (Optional) the default size is 1MB.
    """
    if chunk_size is None:
        chunk_size = DEFAULT_CHUNK_SIZE_BYTES

    logger.debug("Creating temporary directory.")
    with TemporaryDirectory() as tmpdir:
        logger.debug(f"Created temporary directory: {tmpdir}")
        logger.debug("Initiating download request.")
        with requests.get(url, stream=True, timeout=timeout) as r:
            r = cast(requests.Response, r)
            r.raise_for_status()
            logger.debug("Successfully started streaming download.")
            try:
                header = r.headers["Content-Disposition"]
                value, params = cgi.parse_header(header)
                if value != "attachment":
                    raise Exception("Expected header Content-Disposition: attachment")
                filename = params["filename"]
                logger.debug(f"Got filename from Content-Disposition header: {filename}")
            except Exception as e:
                logger.warning("Failed to get filename from headers.", exc_info=e)
                filename = os.path.basename(urlparse(url).path)
                logger.debug(f"Got filename from basename of download URL: {filename}")
            filepath = os.path.join(tmpdir, filename)
            logger.debug(f"Creating file: {filepath}.")
            with open(filepath, "wb") as f:
                logger.debug(f"Writing stream to file: {filepath}")
                for chunk in r.iter_content(chunk_size=chunk_size):
                    f.write(chunk)
        return upload_mesh_from_local_file(
            client,
            project_id,
            filepath,
            name=name,
            scaling=scaling,
            chunk_size=chunk_size,
        )


def upload_mesh(
    client: Client,
    project_id: str,
    path: Union[os.PathLike, str],
    *,
    name: Optional[str] = None,
    scaling: Optional[float] = None,
    chunk_size: Optional[int] = None,
) -> Mesh:
    """
    Upload a mesh from a local file or a URL.

    Parameters
    ----------
    client: Client
        A LuminaryCloud Client (see client.py)
    project_id: str
        The ID of the project to upload the mesh under
    path: PathLike or str
        The URL or file path of the mesh file to upload.
    name: str
        (Optional) the name of the mesh resource on Luminary Cloud. defaults to the filename
        (without the extension).
    chunk_size: int
        (Optional) the default size is 1MB.
    """
    parsed = urlparse(str(path))
    if parsed.scheme in ("file", ""):
        logger.info(f"Attempting to upload mesh from local file: {path}")
        return upload_mesh_from_local_file(
            client,
            project_id,
            path,
            name=name,
            scaling=scaling,
            chunk_size=chunk_size,
        )
    else:
        logger.info(f"Attempting to upload mesh from URL: {path}")
        return upload_mesh_from_url(
            client,
            project_id,
            str(path),
            name=name,
            scaling=scaling,
            chunk_size=chunk_size,
        )

==> luminarycloud-0.4.4-py3-none-any.whl: luminarycloud/_helpers/upload_mesh.py <==
# Copyright 2023-2024 Luminary Cloud, Inc. All Rights Reserved.
import cgi
import io
import logging
import os
import hashlib
import requests
from tempfile import TemporaryDirectory
from typing import cast, Iterator, Tuple, Optional, Union
from urllib.parse import urlparse

from .._proto.common import common_pb2 as commonpb
from .._proto.mesh.mesh_pb2 import Mesh, UploadMeshRequest, UploadMeshResponse
from .._client import Client

logger = logging.getLogger(__name__)

DEFAULT_CHUNK_SIZE_BYTES = 1024 * 1024


def _chunker(
    stream: io.BufferedIOBase,
    chunk_size: int = None,
) -> Iterator[Tuple[int, bytes]]:
    """
    Generator which breaks the stream into chunks.

    Parameters
    ----------
    stream: io.BufferedIOBase
        The stream to chunk
    chunk_size: int
        (Optional) the default chunk size is 1MB

    Yields
    ------
    offset: int
    chunk: bytes
    """
    if chunk_size is None:
        chunk_size = DEFAULT_CHUNK_SIZE_BYTES

    offset = 0
    while True:
        chunk = stream.read(chunk_size)
        bytes_read = len(chunk)
        if bytes_read == 0:
            return
        yield offset, chunk
        if bytes_read < chunk_size:
            return
        offset += bytes_read


def _digest(
    path: Union[os.PathLike, str],
    *,
    chunk_size: int = None,
) -> bytes:
    """
    Computes the sha256 digest of the file at the given path.

    Uses a buffer to avoid loading the entire file into memory.

    Parameters
    ----------
    path: PathLike or str
        The absolute or relative file path
    chunk_size : int
        (Optional) the default is 1MB

    Returns
    -------
    digest: bytes
    """
    with open(path, "rb") as fp:
        m = hashlib.sha256()
        for _, chunk in _chunker(fp, chunk_size):
            m.update(chunk)
    return m.digest()


def _upload_mesh_from_stream(
    client: Client,
    binary_stream: io.BufferedIOBase,
    project_id: str,
    filename: str,
    size: int,
    sha256checksum: bytes,
    *,
    name: Optional[str] = None,
    scaling: Optional[float] = None,
    chunk_size: Optional[int] = None,
) -> UploadMeshResponse:
    """
    Upload mesh from a binary stream.

    Parameters
    ----------
    client: Client
        A LuminaryCloud Client (see client.py)
    binary_stream: io.BufferedIOBase
        A fresh read stream of the mesh file to upload
    project_id: str
        The ID of the project to upload the mesh under
    filename: str
        The filename of the mesh. Must include the appropriate filename extension.
    size: int
        The size of the stream to be read
    sha256checksum: str
        The digest of the mesh file to upload
    name: str
        (Optional) the name of the mesh resource on Luminary Cloud. defaults to the filename
        (without the extension).
    chunk_size: int
        (Optional) the default size is 1MB.
    """

    assert size > 0, "stream size must be positive"
    fparts = filename.split(".", maxsplit=1)
    if len(fparts) < 2:
        logger.error("Filename is missing extension.")
        raise ValueError("Filename is missing an extension.")
    mesh_name, file_ext = fparts
    if name is not None:
        mesh_name = cast(str, name)
    logger.debug(f"Uploading mesh with name '{mesh_name}' and extension '{file_ext}'.")

    logger.debug("Uploading chunks.")
    response = None
    next_upload_token = ""
    file_metadata = commonpb.FileMetadata(
        name=mesh_name, ext=file_ext, size=size, sha256_checksum=sha256checksum
    )
    for offset, chunk in _chunker(binary_stream, chunk_size):
        req = UploadMeshRequest(
            project_id=project_id,
            name=mesh_name,
            upload_token=next_upload_token,
            file_chunk=commonpb.FileChunk(
                metadata=file_metadata,
                bytes=commonpb.ByteChunk(
                    start_offset=offset,
                    data=chunk,
                ),
            ),
        )
        if scaling is not None:
            req.scaling = cast(float, scaling)
        response = client.UploadMesh(req)
        next_upload_token = response.upload_token

    if response is None:
        logger.error("Stream is empty.")
        raise ValueError("Empty stream")

    return cast(UploadMeshResponse, response)


def _is_valid_upload_url(path: Union[os.PathLike, str]) -> bool:
    parsed = urlparse(str(path))
    return parsed.scheme in ("http", "https")


def upload_mesh_from_local_file(
    client: Client,
    project_id: str,
    path: Union[os.PathLike, str],
    *,
    name: Optional[str] = None,
    scaling: Optional[float] = None,
    chunk_size: Optional[int] = None,
) -> Mesh:
    """
    Upload a mesh from a local file.

    The mesh file format is inferred from the filename extension.
    For supported formats, see: https://docs.luminarycloud.com/en/articles/9275233-upload-a-mesh

    Parameters
    ----------
    client: Client
        A LuminaryCloud Client (see client.py)
    project_id: str
        The ID of the project to upload the mesh under
    path: PathLike or str
        The relative or absolute file path of the mesh file to upload.
    name: str
        (Optional) the name of the mesh resource on Luminary Cloud. defaults to the filename
        (without the extension).
    chunk_size: int
        (Optional) the default size is 1MB.
    """
    if not os.path.exists(path):
        logger.error("File not found.")
        raise FileNotFoundError

    with open(path, "rb") as fp:
        response = _upload_mesh_from_stream(
            client,
            fp,
            project_id,
            os.path.basename(path),
            os.path.getsize(path),
            _digest(path),
            name=name,
            scaling=scaling,
            chunk_size=chunk_size,
        )

    return response.mesh


def upload_mesh_from_url(
    client: Client,
    project_id: str,
    url: str,
    *,
    timeout: int = 5,
    name: Optional[str] = None,
    scaling: Optional[float] = None,
    chunk_size: Optional[int] = None,
) -> Mesh:
    """
    Upload a mesh from a URL.

    The mesh file format is inferred from the filename extension.
    For supported formats, see: https://docs.luminarycloud.com/en/articles/9275233-upload-a-mesh

    Parameters
    ----------
    client: Client
        A LuminaryCloud Client (see client.py)
    project_id: str
        The ID of the project to upload the mesh under
    url: str
        The url of the mesh file.
    timeout: int
        (Optional) timeout (in seconds) for the download request.
    name: str
        (Optional) the name of the mesh resource on Luminary Cloud. defaults to the filename
        (without the extension).
    chunk_size: int
        (Optional) the default size is 1MB.
    """
    if chunk_size is None:
        chunk_size = DEFAULT_CHUNK_SIZE_BYTES

    logger.debug("Creating temporary directory.")
    with TemporaryDirectory() as tmpdir:
        logger.debug(f"Created temporary directory: {tmpdir}")
        logger.debug("Initiating download request.")
        with requests.get(url, stream=True, timeout=timeout) as r:
            r = cast(requests.Response, r)
            r.raise_for_status()
            logger.debug("Successfully started streaming download.")
            try:
                header = r.headers["Content-Disposition"]
                value, params = cgi.parse_header(header)
                if value != "attachment":
                    raise Exception("Expected header Content-Disposition: attachment")
                filename = params["filename"]
                logger.debug(f"Got filename from Content-Disposition header: {filename}")
            except Exception as e:
                logger.warning("Failed to get filename from headers.", exc_info=e)
                filename = os.path.basename(urlparse(url).path)
                logger.debug(f"Got filename from basename of download URL: {filename}")
            filepath = os.path.join(tmpdir, filename)
            logger.debug(f"Creating file: {filepath}.")
            with open(filepath, "wb") as f:
                logger.debug(f"Writing stream to file: {filepath}")
                for chunk in r.iter_content(chunk_size=chunk_size):
                    f.write(chunk)
        return upload_mesh_from_local_file(
            client,
            project_id,
            filepath,
            name=name,
            scaling=scaling,
            chunk_size=chunk_size,
        )


def upload_mesh(
    client: Client,
    project_id: str,
    path: Union[os.PathLike, str],
    *,
    name: Optional[str] = None,
    scaling: Optional[float] = None,
    chunk_size: Optional[int] = None,
) -> Mesh:
    """
    Upload a mesh from a local file or a URL.

    Parameters
    ----------
    client: Client
        A LuminaryCloud Client (see client.py)
    project_id: str
        The ID of the project to upload the mesh under
    path: PathLike or str
        The URL or file path of the mesh file to upload.
    name: str
        (Optional) the name of the mesh resource on Luminary Cloud. defaults to the filename
        (without the extension).
    chunk_size: int
        (Optional) the default size is 1MB.
    """
    if _is_valid_upload_url(path):
        logger.info(f"Attempting to upload mesh from URL: {path}")
        return upload_mesh_from_url(
            client,
            project_id,
            str(path),
            name=name,
            scaling=scaling,
            chunk_size=chunk_size,
        )
    else:
        logger.info(f"Attempting to upload mesh from local file: {path}")
        return upload_mesh_from_local_file(
            client,
            project_id,
            path,
            name=name,
            scaling=scaling,
            chunk_size=chunk_size,
        )
